from collections.abc import Callable, Sequence
from functools import partial, update_wrapper
from typing import Literal, overload

import numpy as np
from scipy.special import expit

from ..._types import FloatArrayLike, FloatNDArray
from ._validation import _validate_input_cost_loss_churn


@overload
def make_objective_churn(
    model: Literal['catboost'],
    *,
    accept_rate: float = 0.3,
    clv: float | FloatNDArray = 200,
    incentive_fraction: float | FloatNDArray = 0.05,
    contact_cost: float = 15,
) -> tuple['AECObjectiveChurn', 'AECMetricChurn']: ...


@overload
def make_objective_churn(
    model: Literal['xgboost', 'lightgbm'],
    *,
    accept_rate: float = 0.3,
    clv: float | FloatNDArray = 200,
    incentive_fraction: float | FloatNDArray = 0.05,
    contact_cost: float = 15,
) -> Callable[[FloatNDArray, FloatNDArray], tuple[FloatNDArray, FloatNDArray]]: ...


def make_objective_churn(
    model: Literal['xgboost', 'lightgbm', 'catboost'],
    *,
    accept_rate: float = 0.3,
    clv: float | FloatNDArray = 200,
    incentive_fraction: float | FloatNDArray = 0.05,
    contact_cost: float = 15,
) -> (
    Callable[[FloatNDArray, FloatNDArray], tuple[FloatNDArray, FloatNDArray]]
    | tuple['AECObjectiveChurn', 'AECMetricChurn']
):
    """
    Create an objective function for the Expected Cost measure for customer churn.

    The objective function presumes a situation where identified churners are
    contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer.
    For detailed information, consult the paper [1]_.

    Read more in the :ref:`User Guide <cost_functions>`.

    .. seealso::
        :class:`~empulse.models.B2BoostClassifier` : Uses the instance-specific cost function as objective function.

    Parameters
    ----------
    model : {'xgboost', 'lightgbm', 'catboost'}
        The model for which the objective function is created.

        - 'xgboost' : :class:`xgboost:xgboost.XGBClassifier`
        - 'lightgbm' : :class:`lightgbm:lightgbm.LGBMClassifier`
        - 'catboost' : :class:`catboost.CatBoostClassifier`

    accept_rate : float, default=0.3
        Probability of a customer responding to the retention offer (``0 < accept_rate < 1``).

    clv : float or 1D array, shape=(n_samples), default=200
        If ``float``: constant customer lifetime value per retained customer (``clv > incentive_cost``).
        If ``array``: individualized customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).

    incentive_fraction : float, default=0.05
        Cost of incentive offered to a customer, as a fraction of customer lifetime value
        (``0 < incentive_fraction < 1``).

    contact_cost : float, default=1
        Constant cost of contact (``contact_cost > 0``).

    Returns
    -------
    objective : Callable
        A custom objective function for :class:`xgboost:xgboost.XGBClassifier`.

    Examples
    --------

    .. code-block::  python

        from xgboost import XGBClassifier
        from empulse.metrics import make_objective_churn

        objective = make_objective_churn(model='xgboost')
        clf = XGBClassifier(objective=objective, n_estimators=100, max_depth=3)

    Notes
    -----
    The instance-specific cost function for customer churn is defined as [1]_:

    .. math:: C(s_i) = y_i[s_i(f-\\gamma (1-\\delta )CLV_i] + (1-y_i)[s_i(\\delta CLV_i + f)]

    The measure requires that the churn class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    References
    ----------
    .. [1] Janssens, B., Bogaert, M., Bagué, A., & Van den Poel, D. (2022).
        B2Boost: Instance-dependent profit-driven modelling of B2B churn.
        Annals of Operations Research, 1-27.

    """
    if model == 'xgboost':
        objective: Callable[[FloatNDArray, FloatNDArray], tuple[FloatNDArray, FloatNDArray]] = partial(
            _objective,
            accept_rate=accept_rate,
            clv=clv,
            incentive_fraction=incentive_fraction,
            contact_cost=contact_cost,
        )
        update_wrapper(objective, _objective)
    elif model == 'lightgbm':

        def objective(y_true: FloatNDArray, y_score: FloatNDArray) -> tuple[FloatNDArray, FloatNDArray]:
            """
            Create an objective function for the churn AEC measure.

            Parameters
            ----------
            y_true : np.ndarray
                Ground truth labels.
            y_score : np.ndarray
                Predicted values.

            Returns
            -------
            gradient  : np.ndarray
                Gradient of the objective function.

            hessian : np.ndarray
                Hessian of the objective function.
            """
            return _objective(
                y_true,
                y_score,
                accept_rate=accept_rate,
                clv=clv,
                incentive_fraction=incentive_fraction,
                contact_cost=contact_cost,
            )

    elif model == 'catboost':
        return (
            AECObjectiveChurn(
                accept_rate=accept_rate,
                clv=clv,
                incentive_fraction=incentive_fraction,
                contact_cost=contact_cost,
            ),
            AECMetricChurn(
                accept_rate=accept_rate,
                clv=clv,
                incentive_fraction=incentive_fraction,
                contact_cost=contact_cost,
            ),
        )
    else:
        raise ValueError(f"Expected model to be 'xgboost', 'lightgbm' or 'catboost', got {model} instead.")
    return objective


def _objective(
    y_true: FloatNDArray,
    y_score: FloatNDArray,
    accept_rate: float = 0.3,
    clv: float | FloatNDArray = 200,
    incentive_fraction: float | FloatNDArray = 0.05,
    contact_cost: float = 1,
) -> tuple[FloatNDArray, FloatNDArray]:
    """
    Objective function for XGBoost to maximize the profit of a churn model.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth labels (0 or 1).
    y_score : np.ndarray
        Predicted scores.

    Returns
    -------
    gradient  : np.ndarray
        Gradient of the objective function.

    hessian : np.ndarray
        Hessian of the objective function.
    """
    y_proba = expit(y_score)

    incentive_cost = incentive_fraction * clv
    profits = (
        contact_cost + incentive_cost + y_true * (accept_rate * incentive_cost - incentive_cost - clv * accept_rate)
    )
    gradient = y_proba * (1 - y_proba) * profits
    hessian = np.abs((1 - 2 * y_proba) * gradient)
    return gradient, hessian


class AECObjectiveChurn:
    """AEC churn objective for catboost."""

    def __init__(
        self,
        accept_rate: float = 0.3,
        clv: float | FloatNDArray = 200,
        incentive_fraction: float | FloatNDArray = 0.05,
        contact_cost: float = 1,
    ):
        self.accept_rate = accept_rate
        self.clv = clv
        self.incentive_fraction = incentive_fraction
        self.contact_cost = contact_cost

    def calc_ders_range(
        self, predictions: Sequence[float], targets: FloatNDArray, weights: FloatNDArray
    ) -> list[tuple[float, float]]:
        """
        Compute first and second derivative of the loss function with respect to the predicted value for each object.

        Parameters
        ----------
        predictions : indexed container of floats
            Current predictions for each object.

        targets : indexed container of floats
            Target values you provided with the dataset.

        weights : float, optional (default=None)
            Instance weight.

        Returns
        -------
            der1 : list-like object of float
            der2 : list-like object of float

        """
        # Use weights as a proxy to index the costs
        weights = weights.astype(int)
        clv = self.clv[weights] if isinstance(self.clv, np.ndarray) else self.clv

        y_proba = expit(predictions)
        incentive_cost = self.incentive_fraction * clv
        profits = (
            self.contact_cost
            + incentive_cost
            + targets * (self.accept_rate * incentive_cost - incentive_cost - clv * self.accept_rate)
        )
        gradient = y_proba * (1 - y_proba) * profits
        hessian = np.abs((1 - 2 * y_proba) * gradient)
        return list(zip(-gradient, -hessian, strict=False))


class AECMetricChurn:
    """AEC churn metric for catboost."""

    def __init__(
        self,
        accept_rate: float = 0.3,
        clv: float | FloatNDArray = 200,
        incentive_fraction: float | FloatNDArray = 0.05,
        contact_cost: float = 1,
    ):
        self.accept_rate = accept_rate
        self.clv = clv
        self.incentive_fraction = incentive_fraction
        self.contact_cost = contact_cost

    def is_max_optimal(self) -> bool:
        """Return whether great values of metric are better."""
        return False

    def evaluate(
        self, predictions: Sequence[float], targets: FloatNDArray, weights: FloatNDArray
    ) -> tuple[float, float]:
        """
        Evaluate metric value.

        Parameters
        ----------
        predictions : list of indexed containers (containers with only __len__ and __getitem__ defined) of float
            Vectors of approx labels.

        targets : one dimensional indexed container of float
            Vectors of true labels.

        weights : one dimensional indexed container of float, optional (default=None)
            Weight for each instance.

        Returns
        -------
            weighted error : float
            total weight : float

        """
        # Use weights as a proxy to index the costs
        weights = weights.astype(int)
        clv = self.clv[weights] if isinstance(self.clv, np.ndarray) else self.clv

        y_proba = expit(predictions)
        return expected_cost_loss_churn(
            targets,
            y_proba,
            accept_rate=self.accept_rate,
            clv=clv,
            incentive_fraction=self.incentive_fraction,
            contact_cost=self.contact_cost,
            normalize=True,
            check_input=False,
        ), 1

    def get_final_error(self, error: float, weight: float) -> float:
        """
        Return final value of metric based on error and weight.

        Parameters
        ----------
        error : float
            Sum of errors in all instances.

        weight : float
            Sum of weights of all instances.

        Returns
        -------
        metric value : float

        """
        return error


def expected_cost_loss_churn(
    y_true: FloatArrayLike,
    y_proba: FloatArrayLike,
    *,
    accept_rate: float = 0.3,
    clv: float | FloatArrayLike = 200,
    incentive_fraction: float | FloatArrayLike = 0.05,
    contact_cost: float = 1,
    normalize: bool = False,
    check_input: bool = True,
) -> float:
    """
    Expected cost of a classifier for customer churn.

    The cost function presumes a situation where identified churners are
    contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer.
    For detailed information, consult the paper [1]_.

    .. seealso::
        :class:`~empulse.models.B2BoostClassifier` : Uses the instance-specific cost function as objective function.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_proba : 1D array-like, shape=(n_samples,)
        Target probabilities, should lie between 0 and 1.

    accept_rate : float, default=0.3
        Probability of a customer responding to the retention offer (``0 < accept_rate < 1``).

    clv : float or 1D array-like, shape=(n_samples), default=200
        If ``float``: constant customer lifetime value per retained customer (``clv` > incentive_cost``).
        If ``array``: individualized customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).

    incentive_fraction : float, default=0.05
        Cost of incentive offered to a customer, as a fraction of customer lifetime value
        (``0 < incentive_fraction < 1``).

    contact_cost : float, default=1
        Constant cost of contact (``contact_cost > 0``).

    normalize : bool, default=False
        Normalize the cost by the number of samples.
        If ``True``, return the average expected cost for customer churn.

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empc_cost : float
        Instance-specific cost function according to the EMPC measure.

    Notes
    -----
    The instance-specific cost function for customer churn is defined as [1]_:

    .. math:: C(s_i) = y_i[s_i(f-\\gamma (1-\\delta )CLV_i] + (1-y_i)[s_i(\\delta CLV_i + f)]

    The measure requires that the churn class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    References
    ----------
    .. [1] Janssens, B., Bogaert, M., Bagué, A., & Van den Poel, D. (2022).
        B2Boost: Instance-dependent profit-driven modelling of B2B churn.
        Annals of Operations Research, 1-27.

    """  # noqa: D401
    if check_input:
        y_true, y_proba, clv, incentive_fraction = _validate_input_cost_loss_churn(
            y_true,
            y_proba,
            clv=clv,
            accept_rate=accept_rate,
            incentive_fraction=incentive_fraction,
            contact_cost=contact_cost,
        )
    else:
        y_true = np.asarray(y_true)
        y_proba = np.asarray(y_proba)
        clv = np.asarray(clv)
        incentive_fraction = np.asarray(incentive_fraction)

    incentive_cost = incentive_fraction * clv
    profits = y_proba * (
        contact_cost + incentive_cost + y_true * (accept_rate * incentive_cost - incentive_cost - clv * accept_rate)
    )
    if normalize:
        return float(np.mean(profits))
    return float(np.sum(profits))



import numpy as np

from .._types import FloatArrayLike, FloatNDArray


def classification_threshold(y_true: FloatArrayLike, y_score: FloatArrayLike, customer_threshold: float) -> float:
    """
    Return classification threshold for given customer threshold.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('positive': 1, 'negative': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    customer_threshold : float
        Customer threshold determined by value-driven metric.

    Returns
    -------
    threshold : float
        Classification threshold for given customer threshold.

    Examples
    --------
    >>> from empulse.metrics import classification_threshold
    >>> from empulse.metrics import empc
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> score, threshold = empc(y_true, y_score)
    >>> classification_threshold(y_true, y_score, threshold)
    0.2
    """
    y_true = np.asarray(y_true)
    y_score = np.asarray(y_score)
    confusion_matrix, sorted_indices, duplicated_prediction_indices = _compute_confusion_matrix(y_true, y_score)
    classification_thresholds = np.pad(y_score[sorted_indices], pad_width=(1, 0), constant_values=1)
    classification_thresholds = np.delete(classification_thresholds, duplicated_prediction_indices)  # type: ignore[arg-type]
    customer_thresholds = np.sum(confusion_matrix, axis=0) / y_score.shape[0]
    return float(classification_thresholds[np.argmin(np.abs(customer_thresholds - customer_threshold))])


def _compute_confusion_matrix(
    y_true: FloatNDArray, y_pred: FloatNDArray
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]:
    # sort true labels and predictions by highest to the lowest predicted score
    sorted_indices = y_pred.argsort()[::-1]
    sorted_labels = y_true[sorted_indices]
    sorted_predictions = y_pred[sorted_indices]

    # calculate the TP & FP at each new lead targeted
    true_positives = np.pad(np.cumsum(sorted_labels), pad_width=(1, 0))
    false_positives = np.pad(np.cumsum(sorted_labels == 0), pad_width=(1, 0))

    # merge consecutive equal prediction values
    duplicated_prediction_indices = np.where(np.diff(sorted_predictions) == 0)[0] + 1
    true_positives = np.delete(true_positives, duplicated_prediction_indices)
    false_positives = np.delete(false_positives, duplicated_prediction_indices)

    return np.array([true_positives, false_positives]), sorted_indices, duplicated_prediction_indices


def _compute_prior_class_probabilities(y_true: FloatNDArray) -> tuple[float, float]:
    """Calculate prior class probabilities from target values."""
    positive_class_prob = float(np.mean(y_true))  # pi_0
    negative_class_prob = 1 - positive_class_prob  # pi_1

    return positive_class_prob, negative_class_prob


def _compute_tpr_fpr_diffs(
    true_positive_rates: FloatNDArray, false_positive_rates: FloatNDArray
) -> tuple[FloatNDArray, FloatNDArray]:
    """Calculate differences between subsequent true positive rates and false positive rates."""
    tpr_diff = np.diff(true_positive_rates, axis=0)  # F_0(T_i) - F_0(T_{i-1})
    fpr_diff = np.diff(false_positive_rates, axis=0)  # F_1(T_i) - F_1(T_{i-1})

    return tpr_diff, fpr_diff


def _compute_profits(
    y_true: FloatNDArray, y_pred: FloatNDArray, cost_benefits: FloatNDArray
) -> tuple[FloatNDArray, FloatNDArray]:
    n_samples = y_pred.shape[0]
    confusion_matrix, _, _ = _compute_confusion_matrix(y_true, y_pred)
    profit_matrix = np.dot(confusion_matrix.T, cost_benefits) / n_samples
    customer_thresholds = np.sum(confusion_matrix, axis=0) / n_samples
    return profit_matrix, customer_thresholds



import os
import re
import textwrap
import traceback

import pytest
from sklearn import set_config

# Directory containing the user guide files
GUIDE_DIR = 'docs/guide'

# Adjust GUIDE_DIR if the current working directory is "tests/"
if os.getcwd().endswith('tests'):
    GUIDE_DIR = '../' + GUIDE_DIR

# Regular expression to find code blocks
CODE_BLOCK_RE = re.compile(r'\.\. code-block:: python\n\s*([\s\S]*?)(?=\n\S|$)')


def extract_code_blocks(file_content):
    """Extract code blocks from the file content."""
    return ['    ' + match.group(1) for match in CODE_BLOCK_RE.finditer(file_content)]


def execute_code_blocks(code_blocks):
    """Execute each code block and report any errors."""
    set_config(enable_metadata_routing=False)  # reset the global configuration
    exec_globals = {}  # shared environment for all code blocks
    for code in code_blocks:
        code = textwrap.dedent(code)
        try:
            exec(code, exec_globals)
        except Exception as e:  # noqa: BLE001
            pytest.fail(f'Error executing code block:\n{code}\nError: {e}\n{traceback.format_exc()}')


@pytest.mark.slow
@pytest.mark.parametrize(
    'file_path',
    [os.path.join(root, file) for root, _, files in os.walk(GUIDE_DIR) for file in files if file.endswith('.rst')],
)
def test_code_blocks_in_user_guides(file_path):
    """Test that code blocks in user guide files execute without errors."""
    with open(file_path, encoding='utf-8') as f:
        content = f.read()
    code_blocks = extract_code_blocks(content)
    execute_code_blocks(code_blocks)



import warnings

import numpy as np
from packaging.version import Version
from scipy import stats as st

from ..._types import FloatArrayLike, FloatNDArray
from .._convex_hull import _compute_convex_hull
from ..common import _compute_prior_class_probabilities, _compute_tpr_fpr_diffs
from ._validation import _validate_input_emp, _validate_input_empb

if Version(np.version.version) < Version('2.0.0'):
    np.trapezoid = np.trapz  # type: ignore[attr-defined] # noqa: NPY201


def empc_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    alpha: float = 6,
    beta: float = 14,
    clv: float | FloatArrayLike = 200,
    incentive_cost: float = 10,
    contact_cost: float = 1,
    check_input: bool = True,
) -> float:
    """
    :func:`~empulse.metrics.empc()` but only returning the EMPC score.

    EMPC presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer,
    this fraction is described by a :math:`Beta(\\alpha, \\beta)` distribution.
    As opposed to :func:`~empulse.metrics.empb`, the incentive cost is a fixed value,
    rather than a fraction of the customer lifetime value. For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empc` : to also return the fraction of the customer base
        that should be targeted to maximize profit.

        :func:`~empulse.metrics.mpc_score` : for a deterministic version of this metric.

        :func:`~empulse.metrics.empb_score` : for a similar metric, but with a variable incentive cost.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    alpha : float, default=6
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``alpha > 1``).

    beta : float, default=14
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``beta > 1``).

    clv : float or 1D array-like, shape=(n_samples), default=200
        If ``float``: average customer lifetime value of retained customers (``clv > incentive_cost``).
        If ``array``: customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).

        .. note::
            Passing a CLV array is equivalent to passing a float with the average CLV of that array.

    incentive_cost : float, default=10
        Cost of incentive offered to a customer (``incentive_cost > 0``).

    contact_cost : float, default=1
        Cost of contacting a customer (``contact_cost > 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empc : float
        Expected Maximum Profit Measure for Customer Churn.

    Notes
    -----
    The EMPC is defined as [1]_:

    .. math::

        \\int_\\gamma CLV (\\gamma (1 - \\delta) - \\phi) \\pi_0 F_0(T) - \
        CLV (\\delta + \\phi) \\pi_1 F_1(T) d\\gamma

    The EMPC requires that the churn class is encoded as 0, and it is NOT interchangeable (see [3]_ p37).
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    An equivalent R implementation is available in [2]_.

    References
    ----------
    .. [1] Verbraken, T., Verbeke, W. and Baesens, B. (2013).
        A Novel Profit Maximizing Metric for Measuring Classification
        Performance of Customer Churn Prediction Models. IEEE Transactions on
        Knowledge and Data Engineering, 25(5), 961-973. Available Online:
        http://ieeexplore.ieee.org/iel5/69/6486492/06165289.pdf?arnumber=6165289
    .. [2] Bravo, C. and Vanden Broucke, S. and Verbraken, T. (2019).
        EMP: Expected Maximum Profit Classification Performance Measure.
        R package version 2.0.5. Available Online:
        http://cran.r-project.org/web/packages/EMP/index.html
    .. [3] Verbraken, T. (2013). Business-Oriented Data Analytics:
        Theory and Case Studies. Ph.D. dissertation, Dept. LIRIS, KU Leuven,
        Leuven, Belgium, 2013.

    Examples
    --------
    >>> from empulse.metrics import empc_score
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> empc_score(y_true, y_score)
    23.875593418348124

    Using scorer:

    >>> import numpy as np
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import cross_val_score, StratifiedKFold
    >>> from sklearn.metrics import make_scorer
    >>> from empulse.metrics import empa_score
    >>>
    >>> X, y = make_classification(random_state=42)
    >>> model = LogisticRegression()
    >>> cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
    >>> scorer = make_scorer(
    ...     empc_score,
    ...     response_method='predict_proba',
    ...     clv=300,
    ...     incentive_cost=15,
    ... )
    >>> np.mean(cross_val_score(model, X, y, cv=cv, scoring=scorer))
    42.09000050753503
    """
    return empc(
        y_true,
        y_score,
        alpha=alpha,
        beta=beta,
        clv=clv,
        incentive_cost=incentive_cost,
        contact_cost=contact_cost,
        check_input=check_input,
    )[0]


def empc(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    alpha: float = 6,
    beta: float = 14,
    clv: float | FloatArrayLike = 200,
    incentive_cost: float = 10,
    contact_cost: float = 1,
    check_input: bool = True,
) -> tuple[float, float]:
    """
    Expected Maximum Profit Measure for Customer Churn (EMPC).

    EMPC presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer,
    this fraction is described by a :math:`Beta(\\alpha, \\beta)` distribution.
    As opposed to :func:`~empulse.metrics.empb`, the incentive cost is a fixed value,
    rather than a fraction of the customer lifetime value. For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empc_score` : to only return the EMPC score.

        :func:`~empulse.metrics.mpc` : for a deterministic version of this metric.

        :func:`~empulse.metrics.empb` : for a similar metric, but with a variable incentive cost.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    alpha : float, default=6
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``alpha > 1``).

    beta : float, default=14
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``beta > 1``).

    clv : float or 1D array-like, shape=(n_samples), default=200
        If ``float``: average customer lifetime value of retained customers (``clv > incentive_cost``).
        If ``array``: customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).

        .. note::
            Passing a CLV array is equivalent to passing a float with the average CLV of that array.

    incentive_cost : float, default=10
        Cost of incentive offered to a customer (``incentive_cost > 0``).

    contact_cost : float, default=1
        Cost of contacting a customer (``contact_cost > 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empc : float
        Expected Maximum Profit Measure for Customer Churn

    threshold : float
        Fraction of the customer base that should be targeted to maximize profit

    Notes
    -----
    The EMPC is defined as [1]_:

    .. math::

        \\int_\\gamma CLV (\\gamma (1 - \\delta) - \\phi) \\pi_0 F_0(T) - \
        CLV (\\delta + \\phi) \\pi_1 F_1(T) d\\gamma

    The EMPC requires that the churn class is encoded as 0, and it is NOT interchangeable (see [3]_ p37).
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    An equivalent R implementation is available in [2]_.

    References
    ----------
    .. [1] Verbraken, T., Verbeke, W. and Baesens, B. (2013).
        A Novel Profit Maximizing Metric for Measuring Classification
        Performance of Customer Churn Prediction Models. IEEE Transactions on
        Knowledge and Data Engineering, 25(5), 961-973. Available Online:
        http://ieeexplore.ieee.org/iel5/69/6486492/06165289.pdf?arnumber=6165289
    .. [2] Bravo, C. and Vanden Broucke, S. and Verbraken, T. (2019).
        EMP: Expected Maximum Profit Classification Performance Measure.
        R package version 2.0.5. Available Online:
        http://cran.r-project.org/web/packages/EMP/index.html
    .. [3] Verbraken, T. (2013). Business-Oriented Data Analytics:
        Theory and Case Studies. Ph.D. dissertation, Dept. LIRIS, KU Leuven,
        Leuven, Belgium, 2013.

    Examples
    --------
    >>> from empulse.metrics import empc
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> empc(y_true, y_score)
    (23.875593418348124, 0.8743700763487141)
    """  # noqa: D401
    if check_input:
        y_true, y_score, clv = _validate_input_emp(y_true, y_score, alpha, beta, clv, incentive_cost, contact_cost)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        clv = np.asarray(clv)

    if isinstance(clv, np.ndarray):
        clv = float(np.mean(clv))

    delta = incentive_cost / clv
    phi = contact_cost / clv
    positive_class_prob, negative_class_prob = _compute_prior_class_probabilities(y_true)

    true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)
    tpr_diff, fpr_diff = _compute_tpr_fpr_diffs(true_positive_rates, false_positive_rates)

    tpr_coef = phi * positive_class_prob
    fpr_coef = (delta + phi) * negative_class_prob

    gamma_bounds = _compute_gamma_bounds(tpr_coef, fpr_coef, delta, tpr_diff, fpr_diff, positive_class_prob)
    gamma_cdf_diff = np.diff(st.beta.cdf(gamma_bounds, a=alpha, b=beta))
    gamma_cdf_1_diff = np.diff(st.beta.cdf(gamma_bounds, a=alpha + 1, b=beta))

    cutoff = len(true_positive_rates) - len(gamma_cdf_diff)
    if cutoff > 0:
        true_positive_rates = true_positive_rates[:-cutoff]
        false_positive_rates = false_positive_rates[:-cutoff]
    mean_gamma = st.beta.mean(a=alpha, b=beta)
    temp_1 = mean_gamma * (clv * (1 - delta) * positive_class_prob * true_positive_rates)
    temp_2 = clv * (tpr_coef * true_positive_rates + fpr_coef * false_positive_rates)
    empc = (temp_1 * gamma_cdf_1_diff - temp_2 * gamma_cdf_diff).sum()

    customer_threshold = (
        gamma_cdf_diff * (positive_class_prob * true_positive_rates + negative_class_prob * false_positive_rates)
    ).sum()

    return empc, customer_threshold


def _compute_gamma_bounds(
    tpr_coef: float,
    fpr_coef: float,
    delta: float,
    tpr_diff: FloatNDArray,
    fpr_diff: FloatNDArray,
    positive_class_prob: float,
) -> FloatNDArray:
    """Compute the gamma bounds of the integral."""
    numerator = fpr_coef * fpr_diff + tpr_coef * tpr_diff
    denominator = positive_class_prob * (1 - delta) * tpr_diff
    # ignore division by zero warning
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        gamma_bounds = numerator / denominator
    gamma_bounds = np.append([0], gamma_bounds)
    return np.append(gamma_bounds[gamma_bounds < 1], [1])


def empb_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    clv: FloatArrayLike,
    alpha: float = 6,
    beta: float = 14,
    incentive_fraction: float = 0.05,
    contact_cost: float = 15,
    check_input: bool = True,
) -> float:
    """
    :func:`~empulse.metrics.empb()` but only returning the EMPB score.

    EMPB presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer,
    this fraction is described by a :math:`Beta(\\alpha, \\beta)` distribution.
    As opposed to :func:`~empulse.metrics.empc`, the incentive cost is a fraction of the customer lifetime value,
    rather than a fixed value. For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empb` : to also return the fraction of the customer base
        that should be targeted to maximize profit.

        :func:`~empulse.metrics.empc_score` : for a similar metric, but with a fixed incentive cost.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    alpha : float, default=6
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``alpha > 1``).

    beta : float, default=14
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``beta > 1``).

    clv : float or 1D array-like, shape=(n_samples)
        If ``float``: average customer lifetime value of retained customers.
        If ``array``: customer lifetime value of each customer when retained.

    incentive_fraction : float, default=0.05
        Cost of incentive offered to a customer, as a fraction of customer lifetime value
        (``0 < incentive_fraction < 1``).

    contact_cost : float, default=15
        Cost of contacting a customer (``contact_cost > 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empb : float
        Expected Maximum Profit Measure for B2B Customer Churn

    References
    ----------
    .. [1] Janssens, B., Bogaert, M., Bagué, A., & Van den Poel, D. (2022).
        B2Boost: Instance-dependent profit-driven modelling of B2B churn.
        Annals of Operations Research, 1-27.
    """
    return empb(
        y_true,
        y_score,
        alpha=alpha,
        beta=beta,
        clv=clv,
        contact_cost=contact_cost,
        incentive_fraction=incentive_fraction,
        check_input=check_input,
    )[0]


def empb(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    clv: FloatArrayLike,
    alpha: float = 6,
    beta: float = 14,
    incentive_fraction: float = 0.05,
    contact_cost: float = 15,
    check_input: bool = True,
) -> tuple[float, float]:
    """
    Expected Maximum Profit Measure for B2B Customer Churn (EMPB).

    EMPB presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer,
    this fraction is described by a :math:`Beta(\\alpha, \\beta)` distribution.
    As opposed to :func:`~empulse.metrics.empc`, the incentive cost is a fraction of the customer lifetime value,
    rather than a fixed value. For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empb_score` : to only return the EMPB score.

        :func:`~empulse.metrics.empc` : for a similar metric, but with a fixed incentive cost.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    alpha : float, default=6
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``alpha > 1``).

    beta : float, default=14
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``beta > 1``).

    clv : float or 1D array-like, shape=(n_samples)
        If ``float``: average customer lifetime value of retained customers.
        If ``array``: customer lifetime value of each customer when retained.

    incentive_fraction : float, default=0.05
        Cost of incentive offered to a customer, as a fraction of customer lifetime value
        (``0 < incentive_fraction < 1``).

    contact_cost : float, default=15
        Cost of contacting a customer (``contact_cost > 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empb : float
        Expected Maximum Profit Measure for B2B Customer Churn

    threshold : float
        Fraction of the customer base that should be targeted to maximize profit

    References
    ----------
    .. [1] Janssens, B., Bogaert, M., Bagué, A., & Van den Poel, D. (2022).
        B2Boost: Instance-dependent profit-driven modelling of B2B churn.
        Annals of Operations Research, 1-27.
    """  # noqa: D401
    if check_input:
        y_true, y_score, clv = _validate_input_empb(y_true, y_score, clv, alpha, beta, incentive_fraction, contact_cost)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        clv = np.asarray(clv)
    gamma = alpha / (alpha + beta)

    # Sort by predicted probabilities
    sorted_indices = np.argsort(y_score)[::-1]
    sorted_y_true = y_true[sorted_indices]
    sorted_clv = clv[sorted_indices]

    # Calculate cumulative sums for benefits and costs
    cumulative_benefits = np.cumsum(gamma * ((1 - incentive_fraction) * sorted_clv - contact_cost) * sorted_y_true)
    cumulative_costs = np.cumsum((-contact_cost - incentive_fraction * sorted_clv) * (1 - sorted_y_true))
    cumulative_profits = cumulative_benefits + cumulative_costs

    # Add a zero at the beginning to indicate not contacting anyone
    cumulative_profits = np.insert(cumulative_profits, 0, 0)

    # Find the maximum profit and corresponding threshold
    max_profit_index = np.argmax(cumulative_profits)
    max_profit = cumulative_profits[max_profit_index]
    threshold = max_profit_index / len(y_score)

    return float(max_profit), float(threshold)


def auepc_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    clv: FloatArrayLike,
    alpha: float = 6,
    beta: float = 14,
    incentive_fraction: float = 0.05,
    contact_cost: float = 15,
    normalize: bool = True,
    check_input: bool = True,
) -> float:
    """
    Area Under the Expected Profit Curve (AUEPC).

    Calculate the area under the ratio of the expected profit of the model and the perfect model.
    The expected profit is based on the EMPB's definition of profit.

    AUEPC presumes a situation where identified churners are contacted and offered an incentive to remain customers.
    Only a fraction of churners accepts the incentive offer,
    this fraction is described by a :math:`Beta(\\alpha, \\beta)` distribution.
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empb` : to return the maximum profit and threshold.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('churn': 1, 'no churn': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    alpha : float, default=6
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``alpha > 1``).

    beta : float, default=14
        Shape parameter of the beta distribution of the probability
        that a churner accepts the incentive (``beta > 1``).

    clv : float or 1D array-like, shape=(n_samples)
        If ``float``: average customer lifetime value of retained customers.
        If ``array``: customer lifetime value of each customer when retained.

    incentive_fraction : float, default=0.05
        Cost of incentive offered to a customer, as a fraction of customer lifetime value
        (``0 < incentive_fraction < 1``).

    contact_cost : float, default=15
        Cost of contacting a customer (``contact_cost > 0``).

    normalize : bool, default=True
        Whether to normalize the AUEPC score. If True, the score is 1 when the model is perfect.
        This parameter is only useful if a part of the expected profit curve is negative.

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empb : float
        Expected Maximum Profit Measure for B2B Customer Churn

    threshold : float
        Fraction of the customer base that should be targeted to maximize profit

    References
    ----------
    .. [1] Rahman, S., Janssens, B., Bogaert, M. (2025).
        Profit-Driven Pre-Processing in B2B Customer Churn Modeling using Fairness Techniques.
        Journal of Business Research.
    """
    if check_input:
        y_true, y_score, clv = _validate_input_empb(y_true, y_score, clv, alpha, beta, incentive_fraction, contact_cost)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
        clv = np.asarray(clv)
    if clv.ndim > 1:
        clv = clv[:, 0]

    accept_rate = alpha / (alpha + beta)

    # Calculate the expected profit vector for the perfect model
    perfect_pred_indices = np.argsort(np.where(y_true == 1, 1, -1) * clv)[::-1]
    perfect_targets = y_true[perfect_pred_indices]
    perfect_clv_targets = clv[perfect_pred_indices]

    perfect_benefits = np.cumsum(
        accept_rate * ((1 - incentive_fraction) * perfect_clv_targets - contact_cost) * perfect_targets
    )
    perfect_costs = np.cumsum((-contact_cost - incentive_fraction * perfect_clv_targets) * (1 - perfect_targets))
    perfect_profits = perfect_benefits + perfect_costs

    # Calculate the expected profit vector for the perfect model
    sorted_indices = y_score.argsort()[::-1]
    targets = y_true[sorted_indices]
    clv_targets = clv[sorted_indices]

    benefits = np.cumsum(accept_rate * ((1 - incentive_fraction) * clv_targets - contact_cost) * targets)
    costs = np.cumsum((-contact_cost - incentive_fraction * clv_targets) * (1 - targets))
    profits = benefits + costs

    # Stop at the point where perfect profits become negative
    stop_index: int = np.argmax(perfect_profits < 0) if np.any(perfect_profits < 0) else len(perfect_profits)  # type: ignore[assignment]

    # Calculate the AUEPC
    score = float(np.trapezoid(profits[:stop_index] / perfect_profits[:stop_index], dx=1 / len(profits)))  # type: ignore[attr-defined]
    if normalize:
        score /= (stop_index - 1) / len(profits)
    return score



from .cost_matrix import CostMatrix
from .cost_metric import Cost
from .metric import Metric
from .metric_strategies import MaxProfit, MetricStrategy
from .savings_metric import Savings

__all__ = ['Cost', 'CostMatrix', 'MaxProfit', 'Metric', 'MetricStrategy', 'Savings']



import sys
import warnings
from collections.abc import Callable
from itertools import product
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

import numpy as np
from imblearn.base import BaseSampler
from numpy.random import RandomState
from numpy.typing import ArrayLike, NDArray
from sklearn.utils import _safe_indexing, check_random_state
from sklearn.utils._param_validation import StrOptions

from .._types import IntNDArray, ParameterConstraint
from ..utils._sklearn_compat import ClassifierTags, Tags, type_of_target  # type: ignore
from ._strategies import Strategy, StrategyFn, _independent_weights

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:  # pragma: no cover
    import pandas as pd

    _XT = TypeVar('_XT', NDArray[Any], pd.DataFrame, ArrayLike)
    _YT = TypeVar('_YT', NDArray[Any], pd.Series, ArrayLike)
else:
    _XT = TypeVar('_XT', bound=ArrayLike)
    _YT = TypeVar('_YT', bound=ArrayLike)


class BiasResampler(BaseSampler):  # type: ignore[misc]
    """
    Sampler which resamples instances to remove bias against a subgroup.

    Read more in the :ref:`User Guide <bias_mitigation>`.

    Parameters
    ----------
    strategy : {'statistical parity', 'demographic parity'} or Callable, default='statistical parity'
        Determines how the group weights are computed.
        Group weights determine how much to over or undersample each combination of target and sensitive feature.
        For example, a weight of 2 for the pair (y_true == 1, sensitive_feature == 0) means that the resampled dataset
        should have twice as many instances with y_true == 1 and sensitive_feature == 0
        compared to the original dataset.

        - ``'statistical_parity'`` or ``'demographic parity'``: \
        probability of positive predictions are equal between subgroups of sensitive feature.

        - ``Callable``: function which computes the group weights based on the target and sensitive feature. \
        Callable accepts two arguments: y_true and sensitive_feature and returns the group weights. \
        Group weights are a 2x2 matrix where the rows represent the target variable and the columns represent the \
        sensitive feature. \
        The element at position (i, j) is the weight for the pair (y_true == i, sensitive_feature == j).
    transform_feature : Optional[Callable], default=None
        Function which transforms sensitive_feature before resampling the training data.
        The function takes in the sensitive feature in the form of a numpy.ndarray
        and outputs the transformed sensitive feature as a numpy.ndarray.
        This can be useful if you want to transform a continuous variable to a binary variable at fit time.
    random_state : int or :class:`numpy:numpy.random.RandomState`, optional
        Random number generator seed for reproducibility.

    Attributes
    ----------
    sample_indices_ : numpy.ndarray
        Indices of the samples that were selected.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.samplers import BiasResampler
        from sklearn.datasets import make_classification

        X, y = make_classification()
        high_clv = np.random.randint(0, 2, y.shape)

        sampler = BiasResampler()
        sampler.fit_resample(X, y, sensitive_feature=high_clv)

    Example with passing high-clv indicator through cross-validation:

    .. code-block:: python

        import numpy as np
        from empulse.samplers import BiasResampler
        from imblearn.pipeline import Pipeline
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import cross_val_score

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        high_clv = np.random.randint(0, 2, y.shape)

        pipeline = Pipeline([
            ('sampler', BiasResampler().set_fit_resample_request(sensitive_feature=True)),
            ('model', LogisticRegression())
        ])

        cross_val_score(pipeline, X, y, params={'sensitive_feature': high_clv})

    Example with passing clv through a grid search and dynamically determining high_clv customer based on training data:

    .. code-block:: python

        import numpy as np
        from empulse.samplers import BiasResampler
        from imblearn.pipeline import Pipeline
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import GridSearchCV

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        clv = np.random.rand(y.size)

        def to_high_clv(clv: np.ndarray) -> np.ndarray:
            return (clv > np.median(clv)).astype(np.int8)

        pipeline = Pipeline([
            ('sampler', BiasResampler(
                transform_feature=to_high_clv
            ).set_fit_resample_request(sensitive_feature=True)),
            ('model', LogisticRegression())
        ])
        param_grid = {'model__C': np.logspace(-5, 2, 10)}

        grid_search = GridSearchCV(pipeline, param_grid=param_grid)
        grid_search.fit(X, y, sensitive_feature=clv)

    References
    ----------

    .. [1] Rahman, S., Janssens, B., & Bogaert, M. (2025).
           Profit-driven pre-processing in B2B customer churn modeling using fairness techniques.
           Journal of Business Research, 189, 115159. doi:10.1016/j.jbusres.2024.115159
    """

    _estimator_type: ClassVar[str] = 'sampler'
    _sampling_type: ClassVar[str] = 'bypass'
    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'strategy': [StrOptions({'statistical parity', 'demographic parity'}), callable],
        'transform_feature': [callable, None],
        'random_state': ['random_state'],
    }
    _strategy_mapping: ClassVar[dict[Strategy, StrategyFn]] = {
        'statistical parity': _independent_weights,
        'demographic parity': _independent_weights,
    }

    if TYPE_CHECKING:  # pragma: no cover
        # BaseEstimator should dynamically generate the method signature at runtime
        def set_fit_resample_request(self, sensitive_feature: bool = False) -> Self:  # noqa: D102
            pass

    def __init__(
        self,
        *,
        strategy: StrategyFn | Strategy = 'statistical parity',
        transform_feature: Callable[[NDArray[Any]], IntNDArray] | None = None,
        random_state: RandomState | int | None = None,
    ):
        super().__init__()
        self.strategy = strategy
        self.transform_feature = transform_feature
        self.random_state = random_state

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags = ClassifierTags(multi_class=False)
        tags.sampler_tags.sample_indices = True
        return tags

    def fit_resample(
        self,
        X: _XT,
        y: _YT,
        *,
        sensitive_feature: ArrayLike | None = None,
    ) -> tuple[_XT, _YT]:
        """
        Resample the data according to the strategy.

        Parameters
        ----------
        X : 2D array-like, shape=(n_samples, n_features)
        y : 1D array-like, shape=(n_samples,)
        sensitive_feature : 1D array-like, shape=(n_samples,)
            Sensitive attribute used to determine which instances to resample.

        Returns
        -------
        X : 2D array-like, shape=(n_samples, n_features)
            Resampled training data.
        y : 1D array-like, shape=(n_samples,)
            Resampled target values.
        """
        X, y = super().fit_resample(X, y, sensitive_feature=sensitive_feature)
        X: _XT
        y: _YT
        return X, y

    def _fit_resample(
        self,
        X: NDArray[Any],
        y: IntNDArray,
        *,
        sensitive_feature: ArrayLike | None = None,
    ) -> tuple[NDArray[Any], IntNDArray]:
        """
        Resample the data according to the strategy.

        Parameters
        ----------
        X : 2D array-like, shape=(n_samples, n_features)
        y : 1D array-like, shape=(n_samples,)
        sensitive_feature : 1D array-like, shape=(n_samples,)
            Sensitive attribute used to determine which instances to resample.

        Returns
        -------
        X : 2D array-like, shape=(n_samples, n_features)
            Resampled training data.
        y : 1D array-like, shape=(n_samples,)
            Resampled target values.
        """
        y_type = type_of_target(y, input_name='y', raise_unknown=True)
        if y_type != 'binary':
            raise ValueError(f'Only binary classification is supported. The type of the target is {y_type}.')
        self.classes_: NDArray[np.int64] = np.unique(y)
        if len(self.classes_) == 1:
            self.sample_indices_ = np.arange(len(y))
            return X, y
        y_binarized = np.where(y == self.classes_[1], 1, 0)

        random_state = check_random_state(self.random_state)
        if sensitive_feature is None:
            self.sample_indices_ = np.arange(len(y))
            return X, y
        sensitive_feature = np.asarray(sensitive_feature)

        if self.transform_feature is not None:
            sensitive_feature = self.transform_feature(sensitive_feature)

        strategy = self._strategy_mapping[self.strategy] if isinstance(self.strategy, str) else self.strategy
        class_weights = strategy(y_binarized, sensitive_feature)
        # if class_weights are all 1, no resampling is needed
        if np.allclose(class_weights, np.ones(class_weights.shape)):
            self.sample_indices_ = np.arange(len(y))
            return X, y
        indices = np.empty((0,), dtype=int)

        unique_attr = np.unique(sensitive_feature)
        if len(unique_attr) == 1:
            warnings.warn(
                'sensitive_feature only contains one class, no resampling is performed.',
                UserWarning,
                stacklevel=2,
            )
            self.sample_indices_ = np.arange(len(y))
            return X, y

        # determine the number of samples to be drawn for each class and sensitive_feature value
        for target_class, sensitive_val in product(np.unique(y_binarized), unique_attr):
            sensitive_val = int(sensitive_val)
            idx_class = np.flatnonzero(y_binarized == target_class)
            idx_sensitive_feature = np.flatnonzero(sensitive_feature == sensitive_val)
            idx_class_sensitive = np.intersect1d(idx_class, idx_sensitive_feature)
            n_samples = int(class_weights[target_class, sensitive_val] * len(idx_class_sensitive))
            if n_samples > len(idx_class_sensitive):  # oversampling
                indices = np.concatenate((indices, idx_class_sensitive))
                indices = np.concatenate((
                    indices,
                    random_state.choice(idx_class_sensitive, n_samples - len(idx_class_sensitive), replace=True),
                ))
            else:  # undersampling
                indices = np.concatenate((indices, random_state.choice(idx_class_sensitive, n_samples, replace=False)))

        self.sample_indices_ = indices

        return _safe_indexing(X, indices), _safe_indexing(y, indices)



import sys
from numbers import Real
from typing import Any, ClassVar, TypeVar

import numpy as np
import sympy
from numpy.typing import ArrayLike

from ..._types import FloatArrayLike, ParameterConstraint

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

try:
    from xgboost import XGBClassifier
except ImportError:
    XGBClassifier = TypeVar('XGBClassifier')  # type: ignore[misc, assignment]
try:
    from lightgbm import LGBMClassifier
except ImportError:
    LGBMClassifier = TypeVar('LGBMClassifier')  # type: ignore[misc, assignment]
try:
    from catboost import CatBoostClassifier
except ImportError:
    CatBoostClassifier = TypeVar('CatBoostClassifier')  # type: ignore[misc, assignment]

from ..._common import Parameter
from ...metrics import Cost, CostMatrix, Metric
from .._base import BaseBoostClassifier
from .csboost import CSBoostClassifier


class B2BoostClassifier(CSBoostClassifier):
    """
    Cost-sensitive gradient boosting classifier for B2B customer churn.

    B2BoostClassifier supports :class:`xgboost:xgboost.XGBClassifier`, :class:`lightgbm:lightgbm.LGBMClassifier`
    and :class:`catboost.CatBoostClassifier`.
    By default, it uses XGBoost classifier with default hyperparameters.

    Read more in the :ref:`User Guide <csboost>`.

    Parameters
    ----------
    estimator : :class:`xgboost:xgboost.XGBClassifier`, :class:`lightgbm:lightgbm.LGBMClassifier` \
    or :class:`catboost.CatBoostClassifier`, optional
        XGBoost or LightGBM classifier to be fit with desired hyperparameters.
        If not provided, a XGBoost classifier with default hyperparameters is used.

    accept_rate : float, default=0.3
        Probability of a customer responding to the retention offer (0 < `accept_rate` < 1).
        Is overwritten if another `accept_rate` is passed to the ``fit`` method.

    clv : float or 1D array-like, shape=(n_samples), default=200
        If ``float``: constant customer lifetime value per retained customer (``clv > incentive_cost``).
        If ``array``: individualized customer lifetime value of each customer when retained
        (``mean(clv) > incentive_cost``).
        Is overwritten if another `clv` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    incentive_fraction : float, default=0.05
        Cost of incentive offered to a customer, as a fraction of customer lifetime value
        (``0 < incentive_fraction < 1``).
        Is overwritten if another `incentive_fraction` is passed to the ``fit`` method.

    contact_cost : float, default=1
        Constant cost of contact (``contact_cost > 0``).
        Is overwritten if another `contact_cost` is passed to the ``fit`` method.

    Attributes
    ----------
    classes_ : numpy.ndarray, shape=(n_classes,)
        Unique classes in the target.

    estimator_ : :class:`xgboost:xgboost.XGBClassifier`
        Fitted XGBoost classifier.

    Notes
    -----
    The instance-specific cost function for customer churn is defined as [1]_:

    .. math:: C(s_i) = y_i[s_i(f-\\gamma (1-\\delta )CLV_i] + (1-y_i)[s_i(\\delta CLV_i + f)]

    The measure requires that the churn class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('churn': 1, 'no churn': 0).

    .. seealso::
        :func:`~empulse.metrics.create_objective_churn` : Creates the instance-dependent cost function
        for customer churn.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.models import B2BoostClassifier
        from sklearn.datasets import make_classification

        X, y = make_classification()
        clv = np.random.rand(y.size) * 100

        model = B2BoostClassifier()
        model.fit(X, y, clv=clv, incentive_fraction=0.1)

    .. code-block:: python

        import numpy as np
        from empulse.models import B2BoostClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import cross_val_score
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler

        set_config(enable_metadata_routing=True)

        X, y = make_classification(n_samples=50)
        clv = np.random.rand(y.size) * 100

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('model', B2BoostClassifier(contact_cost=10).set_fit_request(clv=True))
        ])

        cross_val_score(pipeline, X, y, params={'clv': clv})

    .. code-block:: python

        import numpy as np
        from empulse.metrics import empb_score
        from empulse.models import B2BoostClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import GridSearchCV
        from sklearn.metrics import make_scorer
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from xgboost import XGBClassifier

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        clv = np.random.rand(y.size) * 100
        contact_cost = 10

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('model', B2BoostClassifier(
                XGBClassifier(n_jobs=2, n_estimators=10),
                contact_cost=contact_cost
            ).set_fit_request(clv=True))
        ])
        param_grid = {
            'model__estimator__learning_rate': np.logspace(-5, 0, 5),
        }
        scorer = make_scorer(
            empb_score,
            response_method='predict_proba',
            contact_cost=contact_cost
        )
        scorer = scorer.set_score_request(clv=True)

        grid_search = GridSearchCV(pipeline, param_grid=param_grid, scoring=scorer)
        grid_search.fit(X, y, clv=clv)

    References
    ----------
    .. [1] Janssens, B., Bogaert, M., Bagué, A., & Van den Poel, D. (2022).
        B2Boost: Instance-dependent profit-driven modelling of B2B churn.
        Annals of Operations Research, 1-27.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        **BaseBoostClassifier._parameter_constraints,
        'accept_rate': [Real],
        'clv': ['array-like', Real],
        'incentive_fraction': [Real],
        'contact_cost': [Real],
    }

    def __init__(
        self,
        estimator: XGBClassifier | LGBMClassifier | CatBoostClassifier | None = None,
        *,
        accept_rate: float = 0.3,
        clv: float | FloatArrayLike = 200,
        incentive_fraction: float = 0.05,
        contact_cost: float = 15,
    ) -> None:
        lifetime, delta, f, gamma = sympy.symbols('clv delta f gamma')
        cost_matrix = (
            CostMatrix()
            .add_tp_benefit(gamma * (lifetime - lifetime * delta - f))
            .add_tp_benefit((1 - gamma) * -f)
            .add_fp_cost(lifetime * delta + f)
            .alias('accept_rate', gamma)
            .alias('incentive_fraction', delta)
            .alias('contact_cost', f)
            .alias('clv', lifetime)
        )
        loss = Metric(cost_matrix, Cost())
        super().__init__(estimator=estimator, loss=loss)
        self.clv = clv
        self.incentive_fraction = incentive_fraction
        self.contact_cost = contact_cost
        self.accept_rate = accept_rate

    def fit(
        self,
        X: FloatArrayLike,
        y: ArrayLike,
        *,
        accept_rate: float | Parameter = Parameter.UNCHANGED,
        clv: ArrayLike | float | Parameter = Parameter.UNCHANGED,
        incentive_fraction: float | Parameter = Parameter.UNCHANGED,
        contact_cost: float | Parameter = Parameter.UNCHANGED,
        fit_params: dict[str, Any] | None = None,
        **loss_params: Any,
    ) -> Self:
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        y : array-like of shape (n_samples,)

        accept_rate : float, default=0.3
            Probability of a customer responding to the retention offer (``0 < accept_rate < 1``).

        clv : float or 1D array-like, shape=(n_samples), default=200
            If ``float``: constant customer lifetime value per retained customer (``clv > incentive_cost``).
            If ``array``: individualized customer lifetime value of each customer when retained
            (``mean(clv) > incentive_cost``).

        incentive_fraction : float, default=10
            Cost of incentive offered to a customer, as a fraction of customer lifetime value
            (``0 < incentive_fraction < 1``).

        contact_cost : float, default=1
            Constant cost of contact (``contact_cost > 0``).

        fit_params : dict, optional
            Additional parameters to pass to the estimator's fit method.

        loss_params : dict
            Additional keyword arguments to pass to the loss function.

        Returns
        -------
        self : B2BoostClassifier
            Fitted B2Boost model.
        """
        if accept_rate is Parameter.UNCHANGED:
            accept_rate = self.accept_rate
        if clv is Parameter.UNCHANGED:
            clv = self.clv
        if incentive_fraction is Parameter.UNCHANGED:
            incentive_fraction = self.incentive_fraction
        if contact_cost is Parameter.UNCHANGED:
            contact_cost = self.contact_cost
        if not isinstance(clv, float | int):
            clv = np.asarray(clv)

        super().fit(
            X,
            y,
            accept_rate=accept_rate,
            clv=clv,
            incentive_fraction=incentive_fraction,
            contact_cost=contact_cost,
            fit_params=fit_params,
            **loss_params,
        )
        return self



from collections.abc import Callable
from typing import Any, Literal

import numpy as np

from .._types import FloatNDArray, IntNDArray

Strategy = Literal['statistical parity', 'demographic parity']
StrategyFn = Callable[[FloatNDArray, IntNDArray], FloatNDArray]


def _independent_weights(y_true: FloatNDArray, protected_attr: IntNDArray) -> FloatNDArray:
    """
    Compute weights so that y would be statistically independent of the protected attribute.

    Parameters
    ----------
    protected_attr : np.ndarray of shape (n_samples,)
        Protected attribute for each sample.
    y_true : np.ndarray of shape (n_samples,)
        Target variable for each sample.

    Returns
    -------
    group_weights : np.ndarray of shape (n_samples,)
        Weights for each y_true and protected_attr pair.
    """
    prob_protected_attr = protected_attr.mean()
    prob_y = y_true.mean()
    prior_protected_attr = np.array([1 - prob_protected_attr, prob_protected_attr]).reshape((1, 2))
    prior_y = np.array([1 - prob_y, prob_y]).reshape((2, 1))
    if protected_attr.ndim == 2:
        protected_attr = protected_attr.reshape(-1)
    joint_prob = np.histogram2d(y_true, protected_attr, bins=(2, 2))[0] / len(y_true)
    epsilon: np.floating[Any] = np.finfo(float).eps  # to avoid division by zero
    weights: FloatNDArray = (1 / (joint_prob + epsilon)) * prior_protected_attr * prior_y
    return weights



import numbers
from typing import TypeVar

import numpy as np
from numpy.typing import NBitBase

from ..._types import FloatArrayLike, FloatNDArray
from .._validation import _check_fraction, _check_gt_one, _check_positive, _check_shape, _check_y_pred, _check_y_true

T = TypeVar('T', bound=NBitBase)


def _validate_input(
    y_true: FloatArrayLike, y_pred: FloatArrayLike, clv: float | FloatArrayLike, d: float, f: float
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray | float]:
    y_true = _check_y_true(y_true)
    y_pred = _check_y_pred(y_pred)
    _check_shape(y_true, y_pred)
    if not isinstance(clv, numbers.Real):
        clv = np.asarray(clv)
        mean_clv = np.mean(clv)
        _check_positive(float(mean_clv), 'clv')
    else:
        _check_positive(clv, 'clv')
    _check_positive(d, 'incentive_cost')
    _check_positive(f, 'contact_cost')
    if isinstance(clv, numbers.Real) and clv <= d:
        raise ValueError(f'clv should be greater than d, got a value of {clv} for clv and for {d} instead.')
    if isinstance(clv, np.ndarray) and np.mean(clv) <= d:
        raise ValueError(f'mean clv should be greater than d, got a value of {clv} for mean clv and for {d} instead.')

    return y_true, y_pred, clv


def _validate_input_emp(
    y_true: FloatArrayLike,
    y_pred: FloatArrayLike,
    alpha: float,
    beta: float,
    clv: float | FloatArrayLike,
    d: float,
    f: float,
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray | float]:
    _check_gt_one(alpha, 'alpha')
    _check_gt_one(beta, 'beta')
    return _validate_input(y_true, y_pred, clv, d, f)


def _validate_input_mp(
    y_true: FloatArrayLike, y_pred: FloatArrayLike, gamma: float, clv: float | FloatArrayLike, d: float, f: float
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray | float]:
    _check_fraction(gamma, 'gamma')
    return _validate_input(y_true, y_pred, clv, d, f)


def _validate_input_mpc(
    y_true: FloatArrayLike,
    y_pred: FloatArrayLike,
    clv: FloatArrayLike,
    accept_rate: float,
    incentive_fraction: float,
    contact_cost: float,
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray | float]:
    y_true = _check_y_true(y_true)
    y_pred = _check_y_pred(y_pred)
    _check_shape(y_true, y_pred)
    clv = np.asarray(clv)
    _check_fraction(accept_rate, 'accept_rate')
    _check_fraction(incentive_fraction, 'incentive_fraction')
    _check_positive(contact_cost, 'contact_cost')

    return y_true, y_pred, clv


def _validate_input_empb(
    y_true: FloatArrayLike,
    y_pred: FloatArrayLike,
    clv: FloatArrayLike,
    alpha: float,
    beta: float,
    incentive_fraction: float,
    contact_cost: float,
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray]:
    y_true = _check_y_true(y_true)
    y_pred = _check_y_pred(y_pred)
    _check_shape(y_true, y_pred)
    clv = np.asarray(clv)
    _check_positive(alpha, 'alpha')
    _check_positive(beta, 'beta')
    _check_fraction(incentive_fraction, 'incentive_fraction')
    _check_positive(contact_cost, 'contact_cost')

    return y_true, y_pred, clv


def _validate_input_cost_loss_churn(
    y_true: FloatArrayLike,
    y_pred: FloatArrayLike,
    clv: FloatArrayLike,
    accept_rate: float,
    incentive_fraction: float | FloatArrayLike,
    contact_cost: float,
) -> tuple[
    FloatNDArray,
    FloatNDArray,
    FloatNDArray,
    FloatNDArray | float,
]:
    y_true = _check_y_true(y_true)
    y_pred = _check_y_pred(y_pred)
    _check_shape(y_true, y_pred)
    clv = np.asarray(clv)
    _check_fraction(accept_rate, 'accept_rate')
    if isinstance(incentive_fraction, float | int):
        _check_fraction(incentive_fraction, 'incentive_fraction')
    else:
        incentive_fraction = np.asarray(incentive_fraction)
    _check_positive(contact_cost, 'contact_cost')

    return y_true, y_pred, clv, incentive_fraction
